﻿#define __DELAY_BACKWARD_COMPATIBLE__
#define F_CPU 1000000UL

#include <avr/io.h>
#include <util/delay.h>
#include <string.h>

typedef unsigned char byte;
typedef unsigned short ushort;
typedef unsigned int uint;

#define CS			_BV(PORTC5)
#define DCS			_BV(PORTB2)
#define WRX			_BV(PORTC3)
#define RDX			_BV(PORTC4)
#define RESET		_BV(PORTC2)
#define D0			_BV(PORTD0)
#define D1			_BV(PORTD1)
#define D2			_BV(PORTD2)
#define D3			_BV(PORTD3)
#define D4			_BV(PORTD4)
#define D5			_BV(PORTD5)
#define D6			_BV(PORTD6)
#define D7			_BV(PORTD7)
#define BKLED		_BV(PORTB1)
#define SLED		_BV(PORTC0)

#define SELECT_LCD() 			(PORTC &= ~CS)
#define DESELECT_LCD() 		(PORTC |=  CS)

#define SWITCH_CMD() 			(PORTB &= ~DCS)
#define SWITCH_DATA() 		(PORTB |=  DCS)

#define START_WRITE() 		(PORTC &= ~WRX)
#define END_WRITE()				(PORTC |=  WRX)
#define START_READ() 			(PORTC &= ~RDX)
#define END_READ()				(PORTC |=  RDX)

#define CMD_RESET					0x01
#define CMD_READ_DISP_ID	0x04
#define CMD_SLEEP_OUT			0x11
#define CMD_DISPLAY_ON		0x28
#define CMD_DISPLAY_OFF		0x29
#define CMD_COL_SET				0x2A
#define CMD_ROW_SET				0x2B
#define CMD_MEM_WRITE			0x2C
#define CMD_MEM_READ			0x2E
#define CMD_MEM_DATA			0x36
#define CMD_PIXEL_FORMAT	0x3A
#define CMD_IDLE_OFF			0x38
#define CMD_IDLE_ON				0x39

#define TOP_PADDING				3
#define LEFT_PADDING			2
#define SCREEN_WIDTH			128
#define SCREEN_HEIGHT			128

#define RGB(r,g,b)			(((r & 0x1f) << 11) | ((g & 0x1f) << 6) | (b & 0x1f))

void draw_str(const char* str);
void init_driver(void);

void blink_led(int time)
{
	while (3)
	{
		PORTC |= SLED;
		_delay_ms(time);
		PORTC &= ~SLED;
		_delay_ms(time);
	}
}

inline void send_cmd(byte cmd)
{
	SWITCH_CMD();
	PORTD = cmd;
	START_WRITE();
	END_WRITE();
}

inline void send_data(byte data)
{
	SWITCH_DATA();
	PORTD = data;
	START_WRITE();
	END_WRITE();
}

void clear_black() 
{
	send_cmd(CMD_MEM_WRITE);

	for (int i = 0; i < 128; i++)
	{
		for (int j = 0; j < 128; j++)
		{
			send_data(0);
			send_data(0);
		}
	}
}

ushort fcolor = RGB(31, 31, 31), bcolor = 0;

int main()
{
	DDRC = 0b111111;
	DDRD = 0b11111111;
	DDRB = 0b110111;

	PORTD = 0;
	PORTC = 0;

	SELECT_LCD();

	PORTC |= RDX;
	PORTB |= BKLED;

	init_driver();

	send_cmd(CMD_DISPLAY_OFF);

	clear_black();

	fcolor = RGB(31, 31, 31); bcolor = 0;
	draw_str("HELLO, ");

	fcolor = RGB(31, 31, 0); bcolor = 0;
	draw_str("FEIFEI");

	fcolor = RGB(31, 31, 31); bcolor = 0;
	draw_str("!\n\n");

	fcolor = RGB(0, 31, 0); bcolor = 0;
	draw_str("this is a LCD display test.\n\n");

	fcolor = RGB(31, 0, 31); bcolor = RGB(0, 0, 0);
	draw_str("\1\1\1 ");

	fcolor = RGB(31, 0, 0); bcolor = RGB(30, 30, 0);
	draw_str("nice!!!");

	fcolor = RGB(31, 0, 31); bcolor = RGB(0, 0, 0);
	draw_str(" \1\1\1");

	fcolor = RGB(12, 12, 31); bcolor = RGB(0, 0, 0);
	draw_str("\n\n1234567890=1>2");

	while (1) 
	{
		_delay_ms(10000);
	}

	return 0;
}


/****** ST7735 Driver ******/

void init_driver()
{
	PORTC |= RESET;
	_delay_ms(50);
	PORTC &= ~RESET;
	_delay_ms(100);
	PORTC |= RESET;
	_delay_ms(120);

	send_cmd(CMD_RESET);     // Software Reset
	_delay_ms(120);

	send_cmd(CMD_SLEEP_OUT); // Sleep out
	_delay_ms(120);

	send_cmd(CMD_MEM_DATA);  // Memory data access control
	send_data(0xC8);

	send_cmd(CMD_PIXEL_FORMAT);	// 色データの通信モード
	send_data(0x05);			 	 // 8ビットデータバスと16ビットの色データ

	send_cmd(CMD_COL_SET);	 // 列アドレス
	send_data(0x00);				 // 開始 02h = 2
	send_data(0x02);
	send_data(0x00);				 // 終了 81h = 129
	send_data(0x81);

	send_cmd(CMD_ROW_SET); 	 // 行アドレス
	send_data(0x00);			   // 開始 0003h = 3
	send_data(0x03);
	send_data(0x00);				 // 終了 0082h = 130
	send_data(0x82);
}

void set_col_range(short start_col, short end_col)
{
	send_cmd(CMD_COL_SET);
	send_data(start_col >> 8);
	send_data(start_col);
	send_data(end_col >> 8);
	send_data(end_col);
}

void set_row_range(short start_row, short end_row)
{
	send_cmd(CMD_ROW_SET);
	send_data(start_row >> 8);
	send_data(start_row);
	send_data(end_row >> 8);
	send_data(end_row);
}

/****** End of ST7735 Driver ******/


#define FONT_WIDTH			6
#define FONT_HEIGHT			12
#define SCREEN_ROWS			10
#define SCREEN_COLS			16

const byte fc_soh[] = { 0x0, 0x1e, 0x21, 0x21, 0x33, 0x21, 0x21, 0x33, 0x2d, 0x21, 0x1e, 0x0 };
const byte fc_stx[] = { 0x3f, 0x21, 0x1e, 0x1e, 0xc, 0x1e, 0x1e, 0xc, 0x12, 0x1e, 0x21, 0x3f };
const byte fc_alert[] = { 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0, 0x8, 0x8, 0x0 };
const byte fc_dot[] = { 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0 };
const byte fc_common[] = { 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x18, 0x10, 0x20 };

const byte fc_0[] = { 0x1e, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x1e, 0x0 };
const byte fc_1[] = { 0x4, 0xc, 0x1c, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0 };
const byte fc_2[] = { 0x1e, 0x21, 0x21, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x20, 0x3f, 0x0 };
const byte fc_3[] = { 0x1e, 0x21, 0x21, 0x1, 0x1, 0xe, 0x1, 0x1, 0x21, 0x21, 0x1e, 0x0 };
const byte fc_4[] = { 0x2, 0x6, 0xa, 0x12, 0x22, 0x3f, 0x2, 0x2, 0x2, 0x2, 0x2, 0x0 };
const byte fc_5[] = { 0x3f, 0x20, 0x20, 0x20, 0x3e, 0x21, 0x1, 0x1, 0x21, 0x21, 0x1e, 0x0 };
const byte fc_6[] = { 0x1e, 0x21, 0x21, 0x20, 0x3e, 0x21, 0x21, 0x21, 0x21, 0x21, 0x1e, 0x0 };
const byte fc_7[] = { 0x3f, 0x1, 0x2, 0x2, 0x4, 0x4, 0x4, 0x8, 0x8, 0x8, 0x8, 0x0 };
const byte fc_8[] = { 0x1e, 0x21, 0x21, 0x21, 0x21, 0x1e, 0x21, 0x21, 0x21, 0x21, 0x1e, 0x0 };
const byte fc_9[] = { 0x1e, 0x21, 0x21, 0x21, 0x21, 0x1f, 0x1, 0x1, 0x1, 0x21, 0x1e, 0x0 };

const byte fc_colon[] = { 0x0, 0x0, 0x0, 0x0, 0xc, 0xc, 0x0, 0x0, 0xc, 0xc, 0x0, 0x0 };
const byte fc_semi[] = { 0x0, 0x0, 0x0, 0x0, 0xc, 0xc, 0x0, 0x0, 0xc, 0xc, 0x4, 0x8 };

const byte fc_less[] = { 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0 };
const byte fc_equal[] = { 0x0, 0x0, 0x0, 0x0, 0x3e, 0x0, 0x3e, 0x0, 0x0, 0x0, 0x0, 0x0 };
const byte fc_great[] = { 0x0, 0x0, 0x2, 0x4, 0x8, 0x10, 0x20, 0x10, 0x8, 0x4, 0x2, 0x0 };
const byte fc_quest[] = { 0x0, 0x1e, 0x21, 0x21, 0x2, 0x4, 0x8, 0x8, 0x0, 0x8, 0x8, 0x0 };
const byte fc_at[] = { 0xc, 0x12, 0x21, 0x2d, 0x35, 0x35, 0x35, 0x2a, 0x20, 0x11, 0xe, 0x0 };

const byte fc_AA[] = { 0xc, 0x12, 0x12, 0x21, 0x21, 0x21, 0x3f, 0x21, 0x21, 0x21, 0x21, 0x0 };
const byte fc_BB[] = { 0x3e, 0x11, 0x11, 0x11, 0x11, 0x1e, 0x11, 0x11, 0x11, 0x11, 0x3e, 0x0 };
const byte fc_CC[] = { 0xc, 0x12, 0x21, 0x20, 0x20, 0x20, 0x20, 0x20, 0x21, 0x12, 0xc, 0x0 };
const byte fc_DD[] = { 0x3c, 0x12, 0x11, 0x11, 0x11, 0x11, 0x10, 0x11, 0x11, 0x12, 0x3c, 0x0 };
const byte fc_EE[] = { 0x3f, 0x20, 0x20, 0x20, 0x20, 0x3e, 0x20, 0x20, 0x20, 0x20, 0x3f, 0x0 };
const byte fc_FF[] = { 0x3f, 0x20, 0x20, 0x20, 0x20, 0x3e, 0x20, 0x20, 0x20, 0x20, 0x20, 0x0 };
const byte fc_GG[] = { 0xe, 0x11, 0x21, 0x20, 0x20, 0x2e, 0x21, 0x21, 0x21, 0x13, 0xd, 0x0 };
const byte fc_HH[] = { 0x21, 0x21, 0x21, 0x21, 0x21, 0x3f, 0x21, 0x21, 0x21, 0x21, 0x21, 0x0 };
const byte fc_II[] = { 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0 };
const byte fc_JJ[] = { 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x12, 0x12, 0xc, 0x0 };
const byte fc_KK[] = { 0x21, 0x22, 0x22, 0x24, 0x28, 0x2c, 0x34, 0x22, 0x22, 0x21, 0x21, 0x0 };
const byte fc_LL[] = { 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x3f, 0x0 };
const byte fc_MM[] = { 0x22, 0x36, 0x36, 0x36, 0x2a, 0x2a, 0x22, 0x22, 0x22, 0x22, 0x22, 0x0 };
const byte fc_NN[] = { 0x21, 0x31, 0x31, 0x29, 0x29, 0x25, 0x25, 0x23, 0x23, 0x21, 0x21, 0x0 };
const byte fc_OO[] = { 0xc, 0x12, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x12, 0xc, 0x0 };
const byte fc_PP[] = { 0x3c, 0x22, 0x21, 0x21, 0x21, 0x22, 0x3c, 0x20, 0x20, 0x20, 0x20, 0x0 };
const byte fc_QQ[] = { 0xc, 0x12, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x25, 0x13, 0xe, 0x1 };
const byte fc_RR[] = { 0x3c, 0x22, 0x21, 0x21, 0x21, 0x22, 0x3c, 0x24, 0x22, 0x22, 0x21, 0x0 };
const byte fc_SS[] = { 0x1e, 0x21, 0x21, 0x20, 0x10, 0xc, 0x2, 0x1, 0x21, 0x21, 0x1e, 0x0 };
const byte fc_TT[] = { 0x1f, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0 };
const byte fc_UU[] = { 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x1e, 0x0 };
const byte fc_VV[] = { 0x22, 0x22, 0x22, 0x22, 0x14, 0x14, 0x14, 0x8, 0x8, 0x8, 0x0, 0x0 };
const byte fc_WW[] = { 0x15, 0x15, 0x15, 0x15, 0x15, 0x15, 0xa, 0xa, 0xa, 0xa, 0xa, 0x0 };
const byte fc_XX[] = { 0x11, 0x11, 0xa, 0xa, 0x4, 0x4, 0x4, 0xa, 0xa, 0x11, 0x11, 0x0 };
const byte fc_YY[] = { 0x11, 0x11, 0x11, 0x11, 0xa, 0xa, 0x4, 0x4, 0x4, 0x4, 0x4, 0x0 };
const byte fc_ZZ[] = { 0x3f, 0x1, 0x2, 0x2, 0x4, 0x4, 0x8, 0x8, 0x10, 0x10, 0x3f, 0x0 };

const byte fc_a[] = { 0x0, 0x0, 0x0, 0x1c, 0x22, 0x2, 0x1c, 0x22, 0x22, 0x26, 0x1b, 0x0 };
const byte fc_b[] = { 0x20, 0x20, 0x20, 0x2c, 0x32, 0x21, 0x21, 0x21, 0x21, 0x32, 0x2c, 0x0 };
const byte fc_c[] = { 0x0, 0x0, 0x0, 0xe, 0x11, 0x20, 0x20, 0x20, 0x20, 0x11, 0xe, 0x0 };
const byte fc_d[] = { 0x1, 0x1, 0x1, 0xd, 0x13, 0x21, 0x21, 0x21, 0x21, 0x13, 0xd, 0x0 };
const byte fc_e[] = { 0x0, 0x0, 0x0, 0x1e, 0x21, 0x21, 0x3f, 0x20, 0x20, 0x21, 0x1e, 0x0 };
const byte fc_f[] = { 0x6, 0x8, 0x8, 0x3e, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0 };
const byte fc_g[] = { 0x0, 0x0, 0x0, 0x1e, 0x21, 0x21, 0x21, 0x21, 0x1e, 0x1, 0x21, 0x1e };
const byte fc_h[] = { 0x20, 0x20, 0x20, 0x2c, 0x32, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x0 };
const byte fc_i[] = { 0x8, 0x8, 0x0, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0 };
const byte fc_j[] = { 0x4, 0x4, 0x0, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x4, 0x18 };
const byte fc_k[] = { 0x20, 0x20, 0x20, 0x20, 0x21, 0x22, 0x24, 0x28, 0x38, 0x24, 0x22, 0x21 };
const byte fc_l[] = { 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x0 };
const byte fc_m[] = { 0x0, 0x0, 0x0, 0x34, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x0 };
const byte fc_n[] = { 0x0, 0x0, 0x0, 0x3c, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22, 0x22 };
const byte fc_o[] = { 0x0, 0x0, 0x0, 0xc, 0x12, 0x21, 0x21, 0x21, 0x21, 0x12, 0xc, 0x0 };
const byte fc_p[] = { 0x0, 0x0, 0x0, 0x2c, 0x32, 0x21, 0x21, 0x21, 0x21, 0x32, 0x2c, 0x20 };
const byte fc_q[] = { 0x0, 0x0, 0x0, 0xd, 0x13, 0x21, 0x21, 0x21, 0x21, 0x13, 0xd, 0x1 };
const byte fc_r[] = { 0x0, 0x0, 0x0, 0x16, 0x18, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10 };
const byte fc_s[] = { 0x0, 0x0, 0x0, 0x1e, 0x21, 0x20, 0x18, 0x6, 0x1, 0x21, 0x1e, 0x0 };
const byte fc_t[] = { 0x8, 0x8, 0x8, 0x1e, 0x8, 0x8, 0x8, 0x8, 0x8, 0x8, 0x6, 0x0 };
const byte fc_u[] = { 0x0, 0x0, 0x0, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x21, 0x1f };
const byte fc_v[] = { 0x0, 0x0, 0x0, 0x22, 0x22, 0x22, 0x22, 0x14, 0x14, 0x8, 0x8, 0x0 };
const byte fc_w[] = { 0x0, 0x0, 0x0, 0x2a, 0x2a, 0x2a, 0x2a, 0x2a, 0x14, 0x14, 0x14, 0x0 };
const byte fc_x[] = { 0x0, 0x0, 0x0, 0x21, 0x12, 0x12, 0xc, 0xc, 0x12, 0x12, 0x21, 0x0 };
const byte fc_y[] = { 0x0, 0x0, 0x0, 0x21, 0x22, 0x12, 0x14, 0x14, 0x8, 0x8, 0x8, 0x30 };
const byte fc_z[] = { 0x0, 0x0, 0x0, 0x3f, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x3f, 0x0 };

/* 自作ASCII表の一部 */
const byte* fc_tab[] = {
	0, fc_soh, fc_stx, 0, 0, 0, 0, 0, 0, 0,   				// 0~9
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   							// 10~19
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   							// 20~29
	0, 0, 0, fc_alert, 0, 0, 0, 0, 0, 0,   					// 30~39
	0, 0, 0, 0, fc_common, 0, fc_dot, 0, fc_0, fc_1,   				// 40~49
	fc_2, fc_3, fc_4, fc_5, fc_6, fc_7, fc_8, fc_9, fc_colon, fc_semi,   							// 50~59
	fc_great, fc_equal, fc_less, fc_quest, fc_at, fc_AA, fc_BB, fc_CC, fc_DD, fc_EE,    		// 60~69
	fc_FF, fc_GG, fc_HH, fc_II, fc_JJ, fc_KK, fc_LL, fc_MM, fc_NN, fc_OO,   	// 70~79
	fc_PP, fc_QQ, fc_RR, fc_SS, fc_TT, fc_UU, fc_VV, fc_WW, fc_XX, fc_YY,    	// 80~89
	fc_ZZ, 0, 0, 0, 0, 0, 0,  				// 90~96

	fc_a, fc_b, fc_c, fc_d, fc_e, fc_f, fc_g,
	fc_h, fc_i, fc_j, fc_k, fc_l, fc_m, fc_n,
	fc_o, fc_p, fc_q, fc_r, fc_s, fc_t,
	fc_u, fc_v, fc_w, fc_x, fc_y, fc_z
};

ushort col = 0, row = 0;
ushort pos_x = LEFT_PADDING, pos_y = TOP_PADDING;

void draw_char(char ch)
{
	// 画面の右側に達したら改行する
	if ((col > SCREEN_COLS - 1) || ch == 10) 
	{
		col = 0; row++;
		pos_x = LEFT_PADDING; pos_y += FONT_HEIGHT + 2;
	}

	if (ch == 10) return;

  // 文字の出力範囲を設定
	set_col_range(pos_x, pos_x + FONT_WIDTH);
	set_row_range(pos_y, pos_y + FONT_HEIGHT);

  // 文字がASCII表にある位置を探す
	const byte* fc = fc_tab[(ushort)ch];

	if (fc != 0)
	{
		send_cmd(CMD_MEM_WRITE);
		
		for (int y = 0; y < FONT_HEIGHT; y++)
		{
			for (int x = 0; x <= FONT_WIDTH; x++)
			{
				ushort k = (*fc) & (1 << (FONT_WIDTH - x));

				// フォントのピクセルを出力する
				send_data(k ? (fcolor >> 8) : (bcolor >> 8));
				send_data(k ? fcolor : bcolor);
			}

			fc++;
		}
	}

	col++;
	pos_x += FONT_WIDTH + 2;
}

void draw_str(const char* str)
{
	uint len = strlen(str);

	for (int i = 0; i < len; i++)
	{
		draw_char(str[i]);
	}
}
